from colorama import Fore
from loggerric._timestamp import *
from loggerric._log_to_file import *

def prompt(question:str, options:list[str]=[], default:str=None, loop_until_valid:bool=False, case_sensitive:bool=True) -> str | None:
    """
    **Prompts standard I/O and returns the answer.**

    If options are given but not chosen by the user during prompting, return value will be `None`

    *Parameters*:
    - `question` (str): The question appearing in the prompt.
    - `options` (list[str]): Options that the user can pick from during prompting.
    - `default` (str): If options are not `None`, optionally specify a default value.
    - `loop_until_valid` (bool): Loop the prompting until the users answer is in the options parameter.
    - `case_sensitive` (bool): Weather or not the answer (with options) is case sensitive.

    *Example (No Options)*:
    ```python
    answer = prompt(question='Insert your name')
    ```

    *Example (Options)*:
    ```python
    answer = prompt(question="What's the best?", options=['a', 'b', 'c'], default='b', loop_until_valid=True, case_sensitive=False)
    ```
    """
    # If no options ommitted prompt immediately
    if len(options) == 0:
        q = f'{Timestamp.get(internal_call=True)}{Fore.BLUE}{question}: {Fore.CYAN}'
        answer = input(q) or None

        # Log to file
        if LogToFileLevel.PROMPT in LogToFile._active_levels:
            LogToFile._log(escape_ansi(q) + answer)
        
        return answer

    # Format options
    options_formatted = f'{Fore.BLUE} | '.join(Fore.YELLOW + o for o in options)
    
    # Lower case if not case sensitive
    if not case_sensitive:
        options = [o.lower() for o in options]

    # Prompt user
    answer = ''
    q = f'{Timestamp.get(internal_call=True)}{Fore.BLUE}{question} [ {options_formatted}{Fore.BLUE} ]{f" ({Fore.YELLOW}{default}{Fore.BLUE})" if default else ""}:{Fore.CYAN} '
    while (answer if case_sensitive else answer.lower()) not in options:
        answer = input(q)

        # Log to file
        if LogToFileLevel.PROMPT in LogToFile._active_levels:
            LogToFile._log(escape_ansi(q) + answer)

        if not loop_until_valid:
            break

        # If the answer is invalid, tell the user
        if (answer if case_sensitive else str(answer).lower()) not in options:
            message = f'{Timestamp.get(internal_call=True)}{Fore.RED}Invalid Option: "{Fore.YELLOW}{answer}{Fore.RED}"{Fore.RESET}'
            print(message)
            
            if LogToFileLevel.PROMPT in LogToFile._active_levels:
                LogToFile._log(escape_ansi(message))

    # Validate answer
    if len(answer) == 0 and default != None:
        return default
    if (answer if case_sensitive else answer.lower()) in options:
        return answer
    
    # Implementor decides what to happen if user answers "wrong"
    return None


from loggerric._log_to_file import LogToFile, LogToFileLevel
from loggerric._progress_bar import ProgressBar
from loggerric._timestamp import Timestamp
from loggerric._log import Log, LogLevel
from loggerric._prompt import prompt
from loggerric._timer import Timer
from colorama import init

# Expose these functions/classes
__all__ = ['Timestamp', 'LogLevel', 'Log', 'prompt', 'ProgressBar', 'Timer', 'LogToFile', 'LogToFileLevel']

# Initialize Colorama
init(autoreset=True)


from time import perf_counter
from colorama import Fore
from loggerric._timestamp import *
from loggerric._log_to_file import *
import math

class ProgressBar:
    """
    **Push a formatted progressbar to the standard output.**

    *Parameters*:
    - `end_value` (int): The end iteration number.
    - `name` (str): The name of the progressbar.
    - `bar_length` (int): The length of the progressbar.

    *Example*:
    ```python
    end_val = 50
    bar = ProgressBar(end_value=end_val, name='Progress Bar Name', bar_length=50)
    for i in range(1, end_val + 1):
        ...
        bar.update(i)
    ```
    """
    # Dunder: Initialization
    def __init__(self, end_value:int, name:str='Progress', bar_length:int=50):
        # Define class scoped variables
        self.end_value = end_value
        self.name = name
        self.bar_length = bar_length
        self._end_value_digets = len(str(end_value))
        self.start_time = 0
        self.elapsed = 0

    # Internal: Method
    def _format_time(self, seconds:float) -> str:
        # Convert seconds to hours, minutes and seconds
        seconds = int(seconds)
        h, m, s = seconds // 3600, (seconds % 3600) // 60, seconds % 60

        # Format to HH:MM:SS
        return f'{h:02d}:{m:02d}:{s:02d}'

    # Method
    def update(self, current_value:int) -> None:
        """
        **Called when the progress bar needs to update.**

        *Parameters*:
        - `current_value` (int): The current value for the progress bar.
        """
        # If its the first time called, log the start time
        if not self.start_time:
            self.start_time = perf_counter()
        
        # String formatted division 'current/end'
        division = f'{str(current_value).zfill(self._end_value_digets)}/{self.end_value}'

        # Calculate the current progress (scaled to bar length)
        progress = current_value * self.bar_length / self.end_value

        # String formatted progress bar
        bar = f'[{Fore.CYAN}{"#" * (math.floor(progress))}{"_" * (self.bar_length - math.floor(progress))}]'
        
        # String formatted percentage complete
        pct = f'{current_value * 100 / self.end_value:.2f}%'

        # Calculate the ETA
        timer = 'ETA --:--:--'
        self.elapsed = perf_counter() - self.start_time
        if current_value > 0:
            remaining = (self.elapsed / current_value) * (self.end_value - current_value)
            timer = f'ETA {self._format_time(remaining)}'

        # Log the progress bar
        progress_bar = f'{Timestamp.get(internal_call=True)}{Fore.BLUE}{self.name}: {Fore.CYAN}{division} {bar} {pct} {timer}{Fore.RESET}'
        print(progress_bar, end='\n' if current_value == self.end_value else '\r')

        # Log to file
        if LogToFileLevel.PROGRESS_BAR in LogToFile._active_levels and current_value == self.end_value:
            LogToFile._log(escape_ansi(progress_bar))


from colorama import Fore
from datetime import datetime

class Timestamp:
    """
    **Manage global timestamps.**
    """
    _format = '{HH}:{MI}:{SS}.{MS} T+{DH}:{DM}:{DS}.{DN}'
    _enabled = True
    _last_get = None

    @classmethod
    def get(cls, dont_override_diff:bool=False, internal_call:bool=False) -> str:
        """
        **Get the current timestamp.**

        *Parameters*:
        - `dont_override_diff` (bool): Determines whether this call updates the timestamp difference reference.
        - `internal_call` (bool): USED INTERNALLY, adds suffix space, brackets, coloring and returns timestamp even if disabled.

        *Example*:
        ```python
        ts = Timestamp.get(dont_override_diff=False)
        ```
        """
        # Check if a timestamp should be returned
        if not cls._enabled and internal_call:
            return ''

        # Grab the current time
        now = datetime.now()

        # Extract difference components
        if cls._last_get is None:
            dh = dm = ds = dn = 0
        else:
            diff = now - cls._last_get
            total_seconds = diff.total_seconds()
            dh, remainder = divmod(total_seconds, 3600)
            dm, ds = divmod(remainder, 60)
            dn = diff.microseconds // 1000

        # Lookup table for replacement loop
        lookup = {
            '{YY}': f'{now.year:04d}', '{MO}': f'{now.month:02d}', '{DD}': f'{now.day:02d}', '{HH}': f'{now.hour:02d}',
            '{MI}': f'{now.minute:02d}', '{SS}': f'{now.second:02d}', '{MS}': str(now.microsecond)[0:min(3, len(str(now.microsecond)))],
            '{DH}': f'{int(dh):02d}', '{DM}': f'{int(dm):02d}', '{DS}': f'{int(ds):02d}', '{DN}': f'{int(dn):03d}',
        }

        # If this is the first message, set diff to 0
        if cls._last_get == 0 or isinstance(cls._last_get, int):
            lookup['{DH}'] = '00'
            lookup['{DM}'] = '00'
            lookup['{DS}'] = '00'
            lookup['{dm}'] = '000'

        # Format timestamp
        timestamp = cls._format
        for k, v in lookup.items():
            timestamp = timestamp.replace(k, v)

        # Check if the difference reference should be set
        if not dont_override_diff:
            cls._last_get = now

        # Return the timestamp
        if internal_call:
            return f'{Fore.MAGENTA}[{timestamp}] {Fore.RESET}'
        else:
            return timestamp
    
    @classmethod
    def set_format(cls, format:str) -> None:
        """
        **Specify the format of the timestamp.**

        *Parameters*:
        - `format` (str): The format of the timestamp.

        Difference = Since last timestamp (unless `dont_override_diff` is `True` in the `Timestamp.get()` function.)
        
        *Fields*:
        - `{YY}` : Year
        - `{MO}` : Month
        - `{DD}` : Day
        - `{HH}` : Hour
        - `{MI}` : Minute
        - `{SS}` : Second
        - `{MS}` : Milliseconds (3 decimals)
        - `{DH}` : Difference hours
        - `{DM}` : Difference minutes
        - `{DS}` : Difference seconds
        - `{DN}` : Difference milliseconds (3 decimals)

        *Example*:
        ```python
        Timestamp.set_format('{HH}:{MM}:{SS}')
        ```
        """
        cls._format = format

    @classmethod
    def enable(cls) -> None:
        """
        **Enable timestamps.**
        """
        cls._enabled = True
    
    @classmethod
    def disable(cls) -> None:
        """
        **Disable timestamps.**
        """
        cls._enabled = False


from time import perf_counter
from colorama import Fore
from loggerric._timestamp import *
from loggerric._log_to_file import *

class Timer:
    """
    **Time how long a code snippet takes to execute.**

    *Arguments*:
    - `name` (str): Name of the timer.

    *Example*:
    ```python
    with Timer(name='Timer Name'):
        ...
    ```
    """
    # Initialization
    def __init__(self, name:str='Timer'):        
        # Define class scoped variables
        self.name = name
        self.end_time = 0
        self.elapsed = 0

    # Entering 'with' operator
    def __enter__(self):
        # Log the start time in variable, and log to STDIO
        self.start_time = perf_counter()
        message = f'{Timestamp.get(internal_call=True)}{Fore.BLUE}{self.name}: {Fore.CYAN}Started...{Fore.RESET}'
        print(message)

        # Log to file
        if LogToFileLevel.TIMER in LogToFile._active_levels:
            LogToFile._log(escape_ansi(message))
        
    # Exiting 'with' operator
    def __exit__(self, exc_type, exc_val, exc_tb):
        # Log the end time, and calculate the time elapsed
        self.end_time = perf_counter()
        self.elapsed = self.end_time - self.start_time

        # Check if duration should be converted to milliseconds
        is_ms = False
        if self.elapsed < 1:
            is_ms = True
            self.elapsed *= 1_000
        
        # Log the duration
        message = f'{Timestamp.get(internal_call=True)}{Fore.BLUE}{self.name}: {Fore.CYAN}Finished... {Fore.YELLOW}{self.elapsed:,.5f} {"ms" if is_ms else "s"}{Fore.CYAN} Elapsed.{Fore.RESET}'
        print(message)

        # Log to file
        if LogToFileLevel.TIMER in LogToFile._active_levels:
            LogToFile._log(escape_ansi(message))


from colorama import Fore
from loggerric._timestamp import *
from loggerric._log_to_file import *
from enum import Enum, auto
import json

def _apply_highlight(text:str, highlight:str, color:str, hl_color:str=Fore.YELLOW):
    highlighted_text = text

    if isinstance(highlight, str):
        highlighted_text = highlighted_text.replace(highlight, hl_color + highlight + color)
    elif isinstance(highlight, list):
        for hl in highlight:
            highlighted_text = highlighted_text.replace(hl, hl_color + hl + color)

    return highlighted_text

class LogLevel(Enum):
    """
    **Enums used for logging.**
    """
    INFO = auto()
    WARN = auto()
    ERROR = auto()
    DEBUG = auto()

class Log:
    """
    **Contains various logging methods.**
    """
    # Keep track of what should be logged
    _active_levels = { LogLevel.INFO, LogLevel.WARN, LogLevel.ERROR, LogLevel.DEBUG }

    @classmethod
    def info(cls, *content:str, highlight:str|list[str]=None) -> None:
        """
        **Format a message as information.**

        *Parameters*:
        - `*content` (str): The content you want printed.
        - `highlight` (str|list[str]): Text that should be highlighted when printing. (Case Sensitive)

        *Example*:
        ```python
        Log.info('Hello World!', ..., highlight='World')
        ```
        """
        # Log the content
        if LogLevel.INFO in cls._active_levels:
            raw_text = ' '.join([str(c) for c in content])

            # Highlight text
            highlighted_text = _apply_highlight(raw_text, highlight, Fore.GREEN)

            finished_text = f'{Timestamp.get(internal_call=True)}{Fore.GREEN}[i] {highlighted_text}{Fore.RESET}'
            print(finished_text)

            # Log to file
            if LogToFileLevel.INFO in LogToFile._active_levels:
                LogToFile._log(escape_ansi(finished_text))

    @classmethod
    def warn(cls, *content:str, highlight:str|list[str]=None) -> None:
        """
        **Format a message as a warning.**

        *Parameters*:
        - `*content` (str): The content you want printed.
        - `highlight` (str|list[str]): Text that should be highlighted when printing. (Case Sensitive)

        *Example*:
        ```python
        Log.warn('Hello World!', ..., highlight='World')
        ```
        """
        # Log the content
        if LogLevel.WARN in cls._active_levels:
            raw_text = ' '.join([str(c) for c in content])

            # Highlight text
            highlighted_text = _apply_highlight(raw_text, highlight, Fore.YELLOW, Fore.WHITE)

            finished_text = f'{Timestamp.get(internal_call=True)}{Fore.YELLOW}[w] {highlighted_text}{Fore.RESET}'
            print(finished_text)

            # Log to file
            if LogToFileLevel.WARN in LogToFile._active_levels:
                LogToFile._log(escape_ansi(finished_text))

    @classmethod
    def error(cls, *content:str, quit_after_log:bool=False, highlight:str|list[str]=None) -> None:
        """
        **Format a message as an error.**

        *Parameters*:
        - `*content` (str): The content you want printed.
        - `quit_after_log` (bool): Quit after logging the error.
        - `highlight` (str|list[str]): Text that should be highlighted when printing. (Case Sensitive)

        *Example*:
        ```python
        Log.error('Hello World!', ..., quit_after_log=True, highlight='World')
        ```
        """
        # Log the content
        if LogLevel.ERROR in cls._active_levels:
            raw_text = ' '.join([str(c) for c in content])

            # Highlight text
            highlighted_text = _apply_highlight(raw_text, highlight, Fore.RED)

            finished_text = f'{Timestamp.get(internal_call=True)}{Fore.RED}[!] {highlighted_text}{Fore.RESET}'
            print(finished_text)

            # Log to file
            if LogToFileLevel.ERROR in LogToFile._active_levels:
                LogToFile._log(escape_ansi(finished_text))

            if quit_after_log: exit()

    @classmethod
    def debug(cls, *content:str, highlight:str|list[str]=None) -> None:
        """
        **Format a message as a debug message.**

        *Parameters*:
        - `*content` (str): The content you want printed.
        - `highlight` (str|list[str]): Text that should be highlighted when printing. (Case Sensitive)

        *Example*:
        ```python
        Log.debug('Hello World!', ..., highlight='World')
        ```
        """
        # Log the content
        if LogLevel.DEBUG in cls._active_levels:
            raw_text = ' '.join([str(c) for c in content])

            # Highlight text
            highlighted_text = _apply_highlight(raw_text, highlight, Fore.LIGHTBLACK_EX)

            finished_text = f'{Timestamp.get(internal_call=True)}{Fore.LIGHTBLACK_EX}[?] {highlighted_text}{Fore.RESET}'
            print(finished_text)

            # Log to file
            if LogToFileLevel.DEBUG in LogToFile._active_levels:
                LogToFile._log(escape_ansi(finished_text))
    
    @classmethod
    def enable(cls, *levels:LogLevel) -> None:
        """
        **Enable logging methods.**

        *Parameters*:
        - `*levels` (LogLevel): Levels that should be enabled.

        *Example*:
        ```python
        Log.enable(LogLevel.INFO, LogLevel.WARN, ...)
        ```
        """
        cls._active_levels.update(levels)
    
    @classmethod
    def disable(cls, *levels:LogLevel) -> None:
        """
        **Disable logging methods.**

        *Parameters*:
        - `*levels` (LogLevel): Levels that should be disabled.

        *Example*:
        ```python
        Log.disable(LogLevel.INFO, LogLevel.WARN, ...)
        ```
        """
        cls._active_levels.difference_update(levels)
    
    @classmethod
    def pretty_print(cls, data, indent:int=4, depth_level:int=0, inline:bool=False) -> None:
        """
        **Print any variable so they are more readable.**

        Intended use is for dictionaries and arrays, other variables still work.

        *Parameters*:
        - `data` (any): The data you want to pretty print.
        - `indent` (int): The indentation amount for the data.
        - `depth_level` (int): USED INTERNALLY, control what child depth the recursive call is at.
        - `inline` (bool): USED INTERNALLY, keeps track of key/value printing, as to not hop to next line.

        *Example*:
        ```python
        data = {
            'name': 'John Doe',
            'age': 27,
            'skills': ['this', 'and', 'that'],
            'status': None,
            'subdict': { 'source': True, 'the_list': ['English', 'Danish'] }
        }
        Log.pretty_print(data)
        ```
        """
        spacing = ' ' * (indent * depth_level)

        if depth_level == 0:
            timestamp = Timestamp.get(internal_call=True)
            print(timestamp)
            if LogToFileLevel.PRETTY_PRINT in LogToFile._active_levels:
                LogToFile._log(escape_ansi(timestamp))

        prefix_spacing = (spacing if not inline else '')

        # Dictionary
        if isinstance(data, dict):
            if not inline:
                print(spacing + Fore.CYAN + '{')
            else:
                print(Fore.CYAN + '{')
            for key, value in data.items():
                key_spacing = ' ' * (indent * (depth_level + 1))
                print(key_spacing + Fore.YELLOW + str(key) + Fore.RESET + ': ', end='')
                if isinstance(value, (dict, list)):
                    cls.pretty_print(value, indent, depth_level + 1, inline=True)
                else:
                    cls.pretty_print(value, indent, depth_level + 1, inline=True)
            print(spacing + Fore.CYAN + '}')

        # List
        elif isinstance(data, list):
            if not inline:
                print(spacing + Fore.MAGENTA + '[')
            else:
                print(Fore.MAGENTA + '[')
            for item in data:
                cls.pretty_print(item, indent, depth_level + 1, inline=False)
            print(spacing + Fore.MAGENTA + ']')

        # String
        elif isinstance(data, str):
            print(prefix_spacing + Fore.GREEN + f'"{data}"')

        # Number
        elif isinstance(data, (int, float, complex)):
            print(prefix_spacing + Fore.BLUE + str(data))

        # Boolean
        elif isinstance(data, bool):
            print(prefix_spacing + Fore.LIGHTBLUE_EX + str(data))

        # None
        elif data is None:
            print(prefix_spacing + Fore.RED + 'None')

        # Other
        else:
            print(prefix_spacing + Fore.RESET + str(data))
        
        # Log to file
        if LogToFileLevel.PRETTY_PRINT in LogToFile._active_levels:
            formatted_data = ''
            try:
                formatted_data = json.dumps(data, indent=indent)
            except TypeError:
                formatted_data = str(data)
            LogToFile._log(formatted_data)
    
    @classmethod
    def table(cls, headers:list[str], rows:list[tuple], table_name:str=None, highlight_rows:list[int]=[], grayout_rows:list[int]=[]) -> None:
        """
        **Format data into a table and print it.**

        *Parameters*:
        - `headers` (list[str]): The headers displayed in the table.
        - `rows` (list[tuple]): The rows displayed in the table.
        - `table_name` (str): The title name of the table.

        *Example*:
        ```python
        headers = ['Item', 'Stock', 'Price']
        rows = [('Cola', '25', '$2.99'), ('GPU', '3', '$995.00'), ('Feather', '2,500', '$0.29')]
        Log.table(headers, rows, table_name='Store Items')
        ```
        """
        column_lengths = { head: len(head) for head in headers }

        # Make sure row and header sizes match, and collect column sizes
        for row in rows:
            assert len(row) == len(headers), 'Header and row size is not matching!'

            for index, head in enumerate(headers):
                assert isinstance(row[index], str), 'An item in the row is not of type "str" !'
                length = len(row[index])
                if column_lengths[head] < length:
                    column_lengths[head] = length
        
        # Check if the highlights and grayouts overlap
        assert len(set(highlight_rows).intersection(set(grayout_rows))) == 0, 'Highlights and grayout rows have overlapping indicies.'

        # Seperator function
        def seperator() -> None:
            seperators = ['-' * (column_lengths[head] + 6) for head in headers]
            _seperator = Fore.BLUE + 'X' + 'X'.join(seperators) + 'X'
            print(_seperator)
            if LogToFileLevel.TABLE in LogToFile._active_levels:
                LogToFile._log(escape_ansi(_seperator))

        # Timestamp
        timestamp = Timestamp.get(internal_call=True)
        print(timestamp)
        if LogToFileLevel.TABLE in LogToFile._active_levels:
            LogToFile._log(escape_ansi(timestamp))

        # Print the title if passed
        if table_name:
            spacing = sum(column_lengths.values()) + len(headers) * 6 + (len(headers) + 11)
            title = Fore.BLUE + f'/  {Fore.YELLOW}{table_name}{Fore.BLUE}  \\'.center(spacing, '#')
            print(title)
            if LogToFileLevel.PRETTY_PRINT in LogToFile._active_levels:
                LogToFile._log(escape_ansi(title))

        seperator()

        # Print the headers
        formatted_headers = [Fore.YELLOW + head.center(column_lengths[head] + 6) for head in headers]
        finished_headers = Fore.BLUE + '|' + f'{Fore.BLUE}|'.join(formatted_headers) + Fore.BLUE + '|'
        print(finished_headers)
        if LogToFileLevel.PRETTY_PRINT in LogToFile._active_levels:
            LogToFile._log(escape_ansi(finished_headers))

        seperator()

        # Print the rows
        for index, row in enumerate(rows):
            # Space and print the rows
            row_color = Fore.CYAN
            if index in highlight_rows:
                row_color = Fore.GREEN
            elif index in grayout_rows:
                row_color = Fore.LIGHTBLACK_EX

            spaced_row = [row_color + item + ' '*(list(column_lengths.values())[index]-len(item)+5) for index, item in enumerate(row)]
            formatted_row = Fore.BLUE + '| ' + f'{Fore.BLUE}| '.join(spaced_row) + Fore.BLUE + '|'
            print(formatted_row)
            if LogToFileLevel.PRETTY_PRINT in LogToFile._active_levels:
                LogToFile._log(escape_ansi(formatted_row))
        
        seperator()


from enum import Enum, auto
from colorama import Fore
import os, re

def escape_ansi(text:str) -> str:
    """
    **Used to escape ansi (terminal color).**

    *Parameters*:
    - `text` (str): The text including the ansi.

    *Example*:
    ```python
    no_ansi = escape_ansi(text='myTextThatIncludesAnsi')
    ```
    """
    regex = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return regex.sub('', text)

class LogToFileLevel(Enum):
    """
    **Enums used for file logging.**
    """
    INFO = auto()
    WARN = auto()
    ERROR = auto()
    DEBUG = auto()
    PRETTY_PRINT = auto()
    TABLE = auto()
    PROGRESS_BAR = auto()
    TIMER = auto()
    PROMPT = auto()

class LogToFile:
    """
    **Contains methods for the automatic file logging.**
    """
    _active_levels = {
        LogToFileLevel.INFO, LogToFileLevel.WARN, LogToFileLevel.ERROR, LogToFileLevel.DEBUG, LogToFileLevel.PROMPT,
        LogToFileLevel.PRETTY_PRINT, LogToFileLevel.TABLE, LogToFileLevel.PROGRESS_BAR, LogToFileLevel.TIMER,
    }

    _should_dump = False
    _dump_path = os.path.join(os.getenv('HOMEDRIVE'), os.getenv('HOMEPATH'), 'loggerric_log_dump.log')

    @classmethod
    def set_dump_location(cls, full_directory:str, file_name:str) -> None:
        """
        **Set a location for the log file.**

        *Parameters*:
        - `full_directory` (str): Full path to the parent directory of the log file.
        - `file_name` (str): File name for the log, without extension.

        *Example*:
        ```python
        LogToFile.set_dump_location(full_directory='C:/Users', file_name='important_file')
        ```
        """
        assert os.path.exists(full_directory), f'Directory "{full_directory}" does not exist!'
        
        cls._dump_path = os.path.join(full_directory, file_name + '.log')

    @classmethod
    def start_logging(cls) -> None:
        """
        **Start logging to file.**
        """
        cls._should_dump = True

        print(f'{Fore.BLUE}Logging To File: {Fore.GREEN}Started{Fore.RESET}')

    @classmethod
    def stop_logging(cls) -> None:
        """
        **Stop logging to file.**
        """
        cls._should_dump = False

        print(f'{Fore.BLUE}Logging To File: {Fore.RED}Stopped{Fore.RESET}')

    @classmethod
    def enable(cls, levels:LogToFileLevel) -> None:
        """
        **Enable levels to log.**

        *Parameters*:
        - `levels` (LogToFileLevel): The levels to enable.

        *Example*:
        ```python
        LogToFile.enable(LogToFileLevel.WARN, ...)
        ```
        """
        cls._active_levels.update(levels)

    @classmethod
    def disable(cls, levels:LogToFileLevel) -> None:
        """
        **Disable levels to log.**

        *Parameters*:
        - `levels` (LogToFileLevel): The levels to disable.

        *Example*:
        ```python
        LogToFile.disable(LogToFileLevel.TIMER, ...)
        ```
        """
        cls._active_levels.difference_update(levels)

    @classmethod
    def _log(cls, data:str) -> None:
        if data:
            with open(cls._dump_path, 'a+t') as file:
                file.write(str(data) + '\n')